#!/usr/bin/env python

"""Tests for `iccas` package."""
from pathlib import Path

import iccas
import pandas as pd

SAMPLE_DATE = "2020-09-29"
SAMPLE_DATETIME = pd.Timestamp(2020, 9, 29, 11, 0)


def test_get(tmp_path: Path):
    d = iccas.get(cache_dir=tmp_path)
    assert isinstance(d, pd.DataFrame)
    assert isinstance(d.index, pd.DatetimeIndex)


def test_get_by_date(tmp_path: Path):
    d = iccas.get_by_date(SAMPLE_DATE, cache_dir=tmp_path)
    assert isinstance(d, pd.DataFrame)
    assert d.index.name == "age_group"
    assert d.attrs["date"] == SAMPLE_DATETIME


def test_get_by_date_keeping_date_column(tmp_path: Path):
    d = iccas.get_by_date(SAMPLE_DATE, cache_dir=tmp_path, keep_date=True)
    assert isinstance(d, pd.DataFrame)
    assert isinstance(d.index, pd.MultiIndex)
    assert isinstance(d.index.levels[0], pd.DatetimeIndex)
    assert d.index.levels[0].name == "date"
    assert d.index.levels[1].name == "age_group"



from pathlib import Path

import click
import matplotlib.pyplot as plt
from cloup import group, option, option_group
from matplotlib.animation import FFMpegWriter
from pathvalidate import sanitize_filename
from tqdm import tqdm

import iccas as ic


def to_filename(string: str) -> str:
    return str(sanitize_filename(string, platform='universal')).replace(' ', '_')


@group('iccas')
def main():
    plt.style.use('seaborn')


@main.command()
@option_group(
    'Chart options',
    option('--variable', type=click.Choice(['cases', 'deaths']), default='cases'),
    option('--window', type=int, default=7,
           help='Size (in days) of the time window'),
    option('-n/-N', '--normalize/--no-normalize', default=True),
    option('--interpolation', type=click.Choice(['linear', 'pchip']), default='pchip'),
    option('--population / --no-population', 'show_population', default=True,
           help='if --normalize, show the distribution of the italian population '
                'age distribution as a reference'),
    option('--lang', type=click.Choice(['it', 'en']), default='it'),
)
@option_group(
    'Video output options',
    option('-s', '--save', default=False, is_flag=True),
    option('--out-dir', type=click.Path(file_okay=False), default='.'),
    option('-f', '--filename', 'filename_fmt', default=None,
           help='file name as Python format string taking arguments {window} and {step}'),
    option('--fps', type=int, default=15),
    option('--dpi', type=int, default=200),
)
@option_group(
    'Interactive plot options',
    option('--interval', default=50),
    option('--repeat', type=bool, default=False),
    option('--repeat-delay', default=2000),
)
def age_dist_animation(
    variable, show_population, lang,
    window, normalize, interpolation,
    save, out_dir, filename_fmt, fps, dpi,
    interval, repeat, repeat_delay,
):
    ic.set_locale(lang)
    data = (
        ic.get()
        .pipe(ic.fix_monotonicity)
        [['cases', 'deaths']]
    )

    animation = ic.charts.AgeDistributionBarChart(
        data,
        variable=variable,
        normalize=normalize,
        window=window,
        population_distribution=(
            ic.get_population_by_age_group().percentage
            if show_population
            else None
        ),
        resample_kwargs=dict(
            method=interpolation,
        )
    ).animation(
        interval=interval,
        repeat=repeat,
        repeat_delay=repeat_delay,
    )

    if save:
        out_dir = Path(out_dir)
        out_dir.mkdir(exist_ok=True, parents=True)
        if filename_fmt is None:
            title = plt.gca().get_title().lower()
            filename = to_filename(title)
        else:
            filename = filename_fmt.format(window=window)
        out_path = (out_dir / filename).with_suffix('.mp4')
        print(out_path)
        writer = FFMpegWriter(fps=fps, metadata=dict(artist='Me'))
        progress_bar = tqdm(
            desc='Saving the video', total=animation.save_count, unit='frame'
        )
        animation.save(
            str(out_path), writer=writer, dpi=dpi,
            progress_callback=lambda *_: progress_bar.update(),
        )
        print(out_path)
    else:
        plt.show()



import pytest

import iccas
from iccas.queries import FIELDS


# fmt: off
@pytest.mark.parametrize("prefixes, fields, expected", [
    ("m", "cases deaths", ["male_cases", "male_deaths"]),
    ("f", "cases deaths", ["female_cases", "female_deaths"]),
    ("t", "*", FIELDS),
    (
        "mf",
        "cases deaths",
        ["male_cases", "male_deaths", "female_cases", "female_deaths"],
    ),
    ("*", "cases", ["cases", "male_cases", "female_cases"]),
])
# fmt: on
def test_cols(prefixes, fields, expected):
    print("ciao")
    actual = iccas.cols(prefixes, fields)
    assert actual == expected


def test_cols_raises_for_invalid_prefix():
    with pytest.raises(ValueError, match="prefixes"):
        iccas.cols("g")
    with pytest.raises(ValueError, match="duplicates"):
        iccas.cols("mfm")
    with pytest.raises(ValueError, match="empty"):
        iccas.cols("")


def test_cols_raises_for_invalid_fields():
    with pytest.raises(ValueError):
        iccas.cols("*", "counts")
    with pytest.raises(ValueError):
        iccas.cols("*", "cases, deaths")
    with pytest.raises(ValueError, match="empty"):
        iccas.cols("*", "")



__author__ = """Gianluca Gippetto"""
__email__ = "gianluca.gippetto@gmail.com"
__version__ = "0.1.2"

__all__ = (
    "charts",
    "checks",
    "cols",
    "get",
    "get_by_date",
    "get_url",
    "load",
    "load_single_date",
    "only_cases",
    "only_counts",
    "only_deaths",
    "resample",
    "reindex_by_interpolating",
    "fix_monotonicity",
    "age_grouper",
    "aggregate_age_groups",
    "get_unknown_sex_count",
    "get_population_by_age",
    "get_population_by_age_group",
    "running_average",
    "running_count",
    "fatality_rate",
    "language",
    "set_language",
    "set_locale",
    "count_by_period",
    "average_by_period",
    "queries",
)

from .i18n import language, set_language, set_locale

from .loading import (
    load_single_date,
    load,
    get,
    get_by_date,
    get_url,
    get_population_by_age,
    get_population_by_age_group,
)
from .processing import (
    reindex_by_interpolating,
    resample,
    fix_monotonicity,
)
from .queries import (
    age_grouper,
    cols,
    aggregate_age_groups,
    only_cases,
    only_counts,
    only_deaths,
    get_unknown_sex_count,
    running_count,
    running_average,
    count_by_period,
    average_by_period,
    fatality_rate,
)
from . import checks
from . import charts
from . import queries



import datetime
from pathlib import Path
from typing import TypeVar, Union

import pandas as pd

DateLike = Union[str, datetime.date]
PathType = Union[str, Path]
PandasObj = TypeVar("PandasObj", pd.Series, pd.DataFrame)



"""
A simple and crappy i18n library written "for fun" and because neither
gettext and python-i18n were optimal for my use case which requires
reloading of translations when a change is made without restarting the
interpreter/kernel and possibly without reloading the module.
"""
import locale
from collections import ChainMap
from dataclasses import dataclass
from io import StringIO
from itertools import chain
from pathlib import Path
from typing import List, Mapping, Optional, Tuple, Union

import yaml

from iccas.types import PathType

LanguageStrings = Mapping[str, Union[str, Mapping[str, str]]]
TranslationsMap = Mapping[str, LanguageStrings]


class Translation:
    def __init__(self, lang, strings):
        self.lang = lang
        self.strings = strings

    def __getitem__(self, str_id: str):
        return self.strings[str_id]

    def get(self, str_id: str, count: int, few_max: int = 5, varname="count") -> str:
        if count < 0:
            raise ValueError("negative count")

        plural_strings = self.strings[str_id]
        if isinstance(plural_strings, str):
            return plural_strings.format(**{varname: count})

        keys: Union[Tuple[str], Tuple[str, str]]
        if count == 0:
            keys = ("zero", "many")
        elif count == 1:
            keys = ("one", "many")
        elif count <= few_max:
            keys = ("few", "many")
        else:
            keys = ("many",)

        for key in keys:
            if key not in plural_strings:
                continue
            return plural_strings[key].format(**{varname: count})
        else:
            raise KeyError(
                f"pluralization error for key '{str_id}' and count {count}: "
                f"none of keys in {keys} were specified"
            )


class NullTranslation(Translation):
    def __init__(self):
        super().__init__('null', {})

    def __getitem__(self, str_id: str) -> str:
        return str_id

    def get(self, str_id: str, count: int, few_max: int = 5, varname="count") -> str:
        return str_id


@dataclass
class TranslationFile:
    path: Path
    content: dict
    last_modified: int

    def reload_if_modified(self) -> bool:
        last_modified = self.path.stat().st_mtime_ns
        if self.last_modified != last_modified:
            self.content = self.read(self.path)
            self.last_modified = self.last_modified
            return True
        return False

    @staticmethod
    def load(path: PathType) -> "TranslationFile":
        path = Path(path)
        content = TranslationFile.read(path)
        last_modified = path.stat().st_mtime_ns
        return TranslationFile(path=path, content=content, last_modified=last_modified)

    @staticmethod
    def read(path: PathType):
        with open(path, encoding="utf-8") as f:
            return yaml.full_load(f)


class TranslationsManager:
    def __init__(self, *paths):
        self.files = [TranslationFile.load(path) for path in set(paths)]
        self.languages = set(chain.from_iterable(f.content.keys() for f in self.files))
        default_lang = locale.getdefaultlocale()[0].split("_")[0]
        self.current_language = default_lang if default_lang in self.languages else "en"

    def _reload_modified_files(self):
        for file in self.files:
            file.reload_if_modified()

    def injector(self, scoped_translations: Union[None, str, TranslationsMap] = None):
        """
        Decorator that adds a ``lang`` argument to the signature and injects a
        ``strings`` argument of type :class:`Translation`.
        If the translation files are modified, they are automatically reloaded.

        You can also pass extra translations as dict or YAML string to the wrapped
        function. Nonetheless, any modification will require to reload the module
        to take effect.

        Args:
            scoped_translations: yaml string or dictionary
        """
        extra_translations: TranslationsMap
        if isinstance(scoped_translations, dict):
            extra_translations = scoped_translations
        if scoped_translations and isinstance(scoped_translations, str):
            with StringIO(scoped_translations) as s:
                extra_translations = yaml.full_load(s) or {}
        elif scoped_translations is None:
            extra_translations = {}
        else:
            raise TypeError('scoped_translations')

        def decorator(func):
            def wrapper(*args, lang: Optional[str] = None, **kwargs):
                lang = lang or self.current_language
                self._reload_modified_files()
                mappings: List[LanguageStrings] = (
                    [extra_translations[lang]] if lang in extra_translations else []
                )
                mappings += [f.content[lang] for f in self.files if lang in f.content]
                strings = Translation(lang, ChainMap(*mappings))
                return func(*args, strings=strings, **kwargs)

            wrapper.translation_manager = self
            wrapper.wrapped_function = func
            return wrapper

        return decorator



from typing import NamedTuple
from urllib.parse import urljoin

import pytest
import requests
import requests_mock

from iccas.caching import RemoteFolderProxy

BASE_URL = "http://www.whatever.com/path/to/folder/"


class RemoteFile(NamedTuple):
    folder_url: str
    relative_path: str
    content: bytes
    etag: str

    @property
    def url(self):
        return urljoin(self.folder_url, self.relative_path)

    def copy_with(self, content: bytes, etag: str):
        return RemoteFile(self.folder_url, self.relative_path, content, etag)


@pytest.fixture(scope="session")
def sample_file():
    return RemoteFile(
        folder_url=BASE_URL,
        relative_path="subpath/to/file.json",
        content="ciao!".encode("utf-8"),
        etag="ABC",
    )


@pytest.fixture
def remote_folder(tmp_path) -> RemoteFolderProxy:
    return RemoteFolderProxy(BASE_URL, tmp_path)


def get_content_callback(file: RemoteFile):
    # Emulate the server
    def content_callback(req: requests.Request, context):
        sent_etag = req.headers.get("If-None-Match", None)
        if sent_etag == file.etag:
            context.status_code = 304
            return b""
        else:
            context.headers["ETag"] = file.etag
            return file.content

    return content_callback


def setup_mocker(mocker, file: RemoteFile):
    return mocker.get(file.url, content=get_content_callback(file))


def test_file_is_downloaded_if_not_in_cache(remote_folder, sample_file):
    with requests_mock.Mocker() as mocker:
        setup_mocker(mocker, sample_file)
        local_path = remote_folder.get(sample_file.relative_path)
        assert local_path.read_bytes() == sample_file.content


def test_file_is_not_downloaded_if_etag_did_not_change(remote_folder, sample_file):
    with requests_mock.Mocker() as mocker:
        # Download a new file and annotate the creation time
        matcher = setup_mocker(mocker, sample_file)
        local_path = remote_folder.get(sample_file.relative_path)
        mtime_before = local_path.stat().st_mtime

        # Request the same file again
        remote_folder.get(sample_file.relative_path)
        # Check that the etag is sent with the request
        assert matcher.last_request.headers["If-None-Match"] == sample_file.etag
        # Check that the cached file remained untouched
        mtime_after = local_path.stat().st_mtime
        assert mtime_before == mtime_after


def test_cached_file_is_overwritten_if_etag_changed(remote_folder, sample_file):
    with requests_mock.Mocker() as mocker:
        # Download a new file
        setup_mocker(mocker, sample_file)
        remote_folder.get(sample_file.relative_path)
        # Simulate a modification in the remote file
        modified_file = sample_file.copy_with(content=b"new content", etag="newetag")
        matcher = setup_mocker(mocker, modified_file)
        # Request the same file again, expecting cache invalidation
        local_path = remote_folder.get(sample_file.relative_path)
        assert matcher.last_request.headers["If-None-Match"] == sample_file.etag
        assert local_path.read_bytes() == modified_file.content


def test_exception_is_raised_if_connection_error_and_file_not_cached(
    remote_folder, sample_file
):
    with requests_mock.Mocker() as mocker:
        # Request a file not in the cache and simulate a ConnectionError
        mocker.get(sample_file.url, exc=requests.exceptions.ConnectionError)
        with pytest.raises(ConnectionError):
            remote_folder.get(sample_file.relative_path)


def test_warning_is_raised_if_connection_error_and_file_in_cache(
    remote_folder, sample_file
):
    with requests_mock.Mocker() as mocker:
        # Request and download a new file
        setup_mocker(mocker, sample_file)
        remote_folder.get(sample_file.relative_path)

        # Request the same file but simulate a ConnectionError
        mocker.get(sample_file.url, exc=requests.exceptions.ConnectionError)
        with pytest.warns(
            UserWarning, match="Could not check if the remote file was modified"
        ):
            path = remote_folder.get(sample_file.relative_path)
        assert path is not None



from typing import Optional, Tuple, Union

import pandas as pd
import matplotlib.pyplot as plt

import iccas as ic

VARIABLES = {"cases", "deaths"}
DEFAULT_CMAP = "inferno"

PeriodBoundary = Union[str, pd.Timestamp]
Period = Tuple[PeriodBoundary, PeriodBoundary]


def check_variable(variable: str):
    if variable not in {"cases", "deaths"}:
        raise ValueError(f"invalid variable: '{variable}' is not in {VARIABLES}")


def check_age_group_size(size: int):
    if size % 10 != 0 or size <= 0:
        raise ValueError(f"invalid age_group_size: {size}")


def legend(ax: Optional[plt.Axes] = None, **kwargs):
    """ Legend positioned at the right of the axes. """
    ax = ax or plt.gca()
    return ax.legend(loc="upper left", bbox_to_anchor=(1.01, 1), **kwargs)


def resample_if_needed(data: pd.DataFrame, freq: str = '1D', **kwargs):
    if data.index.freq is None or data.index.freq.freqstr != freq:
        return ic.resample(data, freq, **kwargs)
    return data



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

import iccas as ic
from iccas.charts.common import DEFAULT_CMAP, legend
from iccas.i18n import translated
from iccas.i18n.lib import NullTranslation, Translation


def add_labels_to_bars(
    rects: List[plt.Rectangle],
    fmt: str = '{:.1f}',
    ax: Optional[plt.Axes] = None,
    y_offset: int = 7,   # in points
    **kwargs,
) -> Tuple[List[plt.Annotation], Callable[[], None]]:
    """
    Attach a text label above each bar in ``rects``, displaying its height.
    Returns a tuple with the list of labels and a function to update the labels
    (useful for animations)
    """
    ax = ax or plt.gca()
    labels = []
    for rect in rects:
        height = rect.get_height()
        label = ax.annotate(
            fmt.format(height),
            xy=(rect.get_x() + rect.get_width() / 2, height),
            ha='center', va='bottom',
            xytext=(0, y_offset), textcoords="offset points",
            **kwargs,
        )
        labels.append(label)

    def update_labels():
        for label, rect, in zip(labels, rects):
            height = rect.get_height()
            label.set_text(fmt.format(height))
            label.xy = label.xy[0], height
            label.xyann = (0, y_offset)

    return labels, update_labels


@translated()
class AgeDistributionBarChart:
    def __init__(
        self,
        counts: pd.DataFrame,
        variable: str = "cases",
        *,
        strings: Translation = NullTranslation(),
        normalize: bool = True,
        age_group_size: int = 10,
        window: int = 14,
        population_distribution: pd.Series = None,
        ax: plt.Axes = None,
        resample_kwargs: Dict[str, Any] = {},
    ):
        """
        Shows the age distribution of cases/deaths at a given date.
        Usable either to draw a static chart or to generate an animation.

        Args:
            counts:
            variable:
            normalize:
            age_group_size:
            window:
            population_distribution:
            ax:
            lang:
        """
        self.ax = ax = ax or plt.gca()
        s = strings
        data = (
            counts[variable]
            .drop(columns="unknown")
            .pipe(ic.running_average, **resample_kwargs)
            .pipe(ic.aggregate_age_groups, cuts=age_group_size)
        )
        if normalize:
            data = data.divide(data.sum(axis=1), axis=0)
            ax.yaxis.set_major_formatter(
                mpl.ticker.PercentFormatter(xmax=1.0, decimals=0)
            )
        self.data = data

        if normalize and population_distribution is not None:
            population = ic.aggregate_age_groups(
                population_distribution, cuts=age_group_size
            )
            sns.set_color_codes("pastel")
            sns.barplot(
                ax=ax,
                label=s["istat_population_data_label"],
                x=population.index,
                y=population,
                facecolor='b',
                hatch="/",
                edgecolor="white",
            )

        sns.set_color_codes("muted")
        age_groups = data.columns
        label = s[f"{variable}_label"]
        self.bars = ax.bar(
            x=age_groups,
            height=[0] * len(age_groups),
            label=label,
            facecolor='b',
            alpha=0.7,
        )
        self.labels, self.update_labels = add_labels_to_bars(
            self.bars,
            fmt='{:.0%}' if normalize else '{:n}',
            fontsize=10,
            ax=ax,
            color='white',
            bbox=dict(
                boxstyle=mpl.patches.BoxStyle("Round", pad=0.3),
                color='black', alpha=.35,
            ),
        )
        ymax = 1.05 * data.max().max()
        ax.set_ylim(0, ymax)
        ax.set_xticklabels(age_groups)
        ax.set_xlabel(s["age"])
        ax.set_ylabel("")
        ax.grid(False, which='both', axis='x')
        if normalize:
            title = s.get(f"running_{variable}_age_distribution", count=window)
        else:
            title = s.get(f"running_average_{variable}_title", count=window)
        ax.set_title(title, fontsize=14)
        if normalize and population_distribution is not None:
            ax.yaxis.set_major_formatter(
                mpl.ticker.PercentFormatter(xmax=1.0, decimals=0)
            )
            ax.legend()

        self.date_text = ax.text(
            0.5,
            0.93,
            "",
            ha="center",
            va="center",
            transform=ax.transAxes,
            color=(1, 1, 1, 0.9),
            bbox=dict(
                boxstyle=mpl.patches.BoxStyle("Round", pad=0.4),
                color=(0.5, 0.5, 0.5, 0.5),
            ),
            fontsize=14,
            fontweight="semibold",
        )
        self.artists = [
            self.date_text,
            *self.bars,
            *self.labels
        ]

    def show(self, date: pd.Timestamp):
        d = self.data.loc[date]
        if isinstance(d, pd.DataFrame):
            if len(d) > 1:
                raise ValueError(f'"{date}" matches more than one date')
            d = d.stack("age_group")
        for bar, h in zip(self.bars, d):
            bar.set_height(h)
        formatted_date = date.strftime("%d %B %Y")
        self.date_text.set_text(formatted_date)
        self.update_labels()
        return self.artists

    def animation(self, **kwargs) -> mpl.animation.FuncAnimation:
        kwargs = {"blit": True, "repeat": False, "interval": 50, **kwargs}
        return mpl.animation.FuncAnimation(
            self.ax.figure, self.show, frames=self.data.index, **kwargs
        )


def age_dist_bar_chart(
    counts: pd.DataFrame,
    date: pd.Timestamp,
    variable: str = "cases",
    *,
    normalize: bool = True,
    age_group_size: int = 10,
    window: int = 14,
    population_distribution: pd.Series = None,
    ax: plt.Axes = None,
    lang: Optional[str] = None,
):
    if ax is None:
        fig, ax = plt.subplots(figsize=(10, 7))
    AgeDistributionBarChart(
        ax=ax,
        counts=counts,
        variable=variable,
        normalize=normalize,
        age_group_size=age_group_size,
        window=window,
        population_distribution=population_distribution,
        lang=lang,    # type: ignore
    ).show(date)


@translated()
def average_by_period_bar_chart(
    counts: pd.DataFrame,
    variable: str,
    *,
    strings: Translation,
    freq: Union[str, int] = 7,
    normalize: bool = False,
    age_group_size: int = 20,
    stacked: bool = True,
    ylim: float = None,
    ax: Optional[plt.Axes] = None,
    figsize: Tuple[float, float] = (12, 7),
) -> plt.Axes:
    if ax is None:
        _, ax = plt.subplots(figsize=figsize)

    # Data preparation
    d = counts[variable].drop(columns='unknown')
    d = ic.aggregate_age_groups(d, age_group_size)
    d = ic.count_by_period(d, freq=freq)
    if normalize:
        d = d.div(d.sum(axis=1), axis=0)

    # Plot
    ax = d.plot.bar(stacked=stacked, cmap=DEFAULT_CMAP, ax=ax, width=.85)
    legend(ax=ax, title=strings['age'])

    # Axes setup
    if freq == 'M':
        xlabel = strings['month']
        date_fmt = '%B'
        xtick_rotation = 0
    else:
        xlabel = strings['period_end_date']
        date_fmt = '%d %b'
        xtick_rotation = 45
    ax.set_xticklabels(d.index.strftime(date_fmt), rotation=xtick_rotation)
    ax.set_xlabel(xlabel)
    if normalize:
        if stacked:
            ax.set_ylim([0, 1.0])
        ax.yaxis.set_major_formatter(mpl.ticker.PercentFormatter(xmax=1.0, decimals=0))
    else:
        ax.yaxis.set_major_formatter(mpl.ticker.StrMethodFormatter('{x:n}'))
        if ylim:
            ax.set_ylim(0, ylim * ax.get_ylim()[1])

    # Title
    base_title = strings[
        f'{variable}_age_distribution' if normalize else f'average_daily_{variable}'
    ]
    if freq == 'M':
        period_label = strings['month_by_month']
    elif freq == 7 or (isinstance(freq, str) and freq.startswith('W')):
        period_label = strings['week_by_week']
    elif isinstance(freq, int):
        period_label = strings.get('by_periods_of_n_days', count=freq)
    title = f"{base_title} {period_label}"
    ax.set_title(title)

    return ax


if __name__ == '__main__':
    plt.style.use('seaborn')
    ic.set_locale('it')
    df = ic.fix_monotonicity(ic.get())
    anim = AgeDistributionBarChart(
        df, population_distribution=ic.get_population_by_age_group().percentage
    ).animation()
    plt.show()



import re
from typing import Union

import numpy as np
import pandas as pd
from pandas.core.dtypes.common import is_integer_dtype

from iccas.queries import cols, only_counts
from iccas.types import PandasObj


def reindex_by_interpolating(
    data: PandasObj,
    new_index: pd.DatetimeIndex,
    preserve_ints: bool = True,
    method="pchip",
    **interpolation,
) -> PandasObj:
    """
    Reindexes `data` and fills new values by interpolation (PCHIP, by default).

    This function was motivated by the fact that :meth:`pandas.DataFrame.resample`
    followed by :meth:`pandas.DataFrame.resample` doesn't take into account
    misaligned datetimes.

    Args:
        data:
            a DataFrame or Series with a datetime index
        new_index:
        preserve_ints:
            after interpolation, columns containing integers in the original
            dataframe are rounded and converted back to int
        method:
            interpolation method (see :meth:`pandas.DataFrame.interpolate`)
        **interpolation:
            other interpolation keyword argument different from `method` passed
            to :meth:`pandas.DataFrame.interpolate`

    Returns:
        a new Dataframe/Series

    See Also:
        :func:`reindex_by_interpolating`
    """
    extended_index = data.index.union(new_index)
    out = data.reindex(extended_index)
    out = out.interpolate(method=method, **interpolation)  # fill NaNs
    out = out.loc[new_index]
    if preserve_ints:
        if isinstance(data, pd.DataFrame):
            int_cols = [
                col for col, dtype in data.dtypes.items() if is_integer_dtype(dtype)
            ]
            out.loc[:, int_cols] = out.loc[:, int_cols].round().astype(int)
        elif is_integer_dtype(data):
            out = out.round().astype(int)

    out.index.rename("date", inplace=True)
    return out


def resample(
    data: PandasObj,
    freq: Union[int, str] = "1D",
    hour: int = 18,
    preserve_ints: bool = True,
    method="pchip",
    **interpolation,
) -> PandasObj:
    """
    Resamples `data` and fills missing values by interpolation.

    The resulting index is a `pandas.DatetimeIndex` whose elements are spaced by
    accordingly to `freq` and having the time set to `{hour}:00`.

    In the case of "day frequencies" ('{num}D'), the index always includes the
    latest date (`data.index[-1]`): the new index is a datetime range built
    going backwards from the latest date.

    This function was motivated by the fact that :meth:`pandas.DataFrame.resample`
    followed by :meth:`pandas.DataFrame.resample` doesn't take into account
    misaligned datetimes. If you want to back-fill or forward-fill, just use
    :meth:`DataFrame.resample`.

    Args:
        data:
            a DataFrame or Series with a datetime index
        freq:
            resampling frequency in `pandas` notation
        hour:
            reference hour; all datetimes in the new index will have this hour
        preserve_ints:
            after interpolation, columns containing integers in the original
            dataframe are rounded and converted back to int
        method:
            interpolation method (see :meth:`pandas.DataFrame.interpolate`)
        **interpolation:
            other interpolation keyword argument different from `method` passed
            to :meth:`pandas.DataFrame.interpolate`

    Returns:
        a new Dataframe/Series with index elements spaced according to ``freq``

    See Also:
        :func:`reindex_by_interpolating`
    """
    if isinstance(freq, int):
        freq = f"{freq}D"
    else:
        freq = freq.upper()

    match = re.fullmatch(r"(\d*)D", freq)
    if match:
        new_index = pd.date_range(
            data.index[-1].replace(hour=hour, minute=0),
            data.index[0].replace(hour=hour, minute=0),
            freq="-1D" if freq == "D" else f"-{freq}",
        )[::-1]
    else:
        new_index = pd.date_range(
            data.index[0],
            data.index[-1].replace(hour=hour, minute=0),
            freq=freq,
        )
    return reindex_by_interpolating(
        data, new_index, preserve_ints=preserve_ints, method=method, **interpolation
    )


# ===========================
#   DATA CORRECTION
# ===========================
def nullify_series_local_bumps(series: pd.Series):
    """ Set to NaN all elements s[i] such that s[i] > s[i+k] """
    curr_min = np.inf
    bad = set()
    for i in reversed(series.index):
        if series[i] > curr_min:
            bad.add(i)
        else:
            curr_min = series[i]
    return series.mask(series.index.isin(bad))


def nullify_local_bumps(df: pd.DataFrame):
    def f(series):
        if "unknown" in series.name:
            return series
        return nullify_series_local_bumps(series)

    return df.apply(f)


def fix_monotonicity(data: pd.DataFrame, method="pchip", **interpolation):
    """
    Replaces tracts of "cases" and "deaths" time series that break the monotonicity
    of the series with interpolated data, ensuring that the sum of male and female
    counts are less or equal to the total count.

    Args:
        data: a DataFrame containing all integer columns about cases and deaths
        method: interpolation method

    Returns:

    """
    # Fix all individual series independently
    orig = only_counts(data)
    fixed = (
        nullify_local_bumps(orig)
        .interpolate(method=method, **interpolation)
        .round()
        .astype(int)
    )

    # Ensure that (males + females <= total) for each age group (including
    # "unknown age" group) taking the maximum between the fixed totals and the
    # fixed males+females
    for variable in ["deaths", "cases"]:
        males_plus_females = fixed[cols("mf", variable)].sum(axis=1, level=1)
        totals = fixed[variable]
        totals_fixed = (
            totals
            .where(totals >= males_plus_females)
            .fillna(males_plus_females)
        )
        fixed[variable] = totals_fixed

    return fixed.astype(int)



"""
Sanity checks.
"""


def is_non_decreasing(df):
    deltas = df.drop(columns="unknown", level=1).diff(1).dropna()
    return (deltas >= 0).all(axis=None)


def totals_not_less_than_sum_of_sexes(data, variable):
    assert variable in {"cases", "deaths"}
    total = data[variable]
    sum_of_sexes = data[f"male_{variable}"] + data[f"female_{variable}"]
    return (total - sum_of_sexes >= 0).all(axis=None)



import locale
from contextlib import contextmanager
from pathlib import Path

from iccas.i18n.lib import TranslationsManager

# These should work both on Windows and Linux (provided that the corresponding
# language packs are installed, e.g. language-pack-it)
LANG_TO_LOCALE = {
    'it': 'it_IT.UTF-8',
    'en': 'en_US.UTF-8'
}

_TRANSLATIONS_PATH = Path(__file__).parent / "translations.yml"
TRANSLATIONS = TranslationsManager(_TRANSLATIONS_PATH)

# Decorator for injecting global translations + optional scoped translations
translated = TRANSLATIONS.injector


def set_language(lang):
    """
    Sets the language. Supported languages: Italian ("it") and English ("en")
    """
    TRANSLATIONS.current_language = lang


@contextmanager
def language(lang):
    entry_lang = TRANSLATIONS.current_language
    TRANSLATIONS.current_language = lang
    yield
    TRANSLATIONS.current_language = entry_lang


def set_locale(lang):
    """ Apart from setting the internal language of the package, also sets the
    locale accordingly so that pandas/matplotlib displays translated dates """
    if lang not in TRANSLATIONS.languages:
        raise ValueError(f'language not supported: {lang}')
    locale.setlocale(locale.LC_ALL, LANG_TO_LOCALE[lang])
    set_language(lang)



import pandas as pd
import pytest

import iccas
from iccas.processing import nullify_series_local_bumps

# A replacement for NaN such that (NA == NA)
NA = -1


@pytest.fixture(scope="session")
def _data(tmpdir_factory):
    cache_dir = tmpdir_factory.mktemp("data")
    return iccas.get(cache_dir=cache_dir).head(10)


@pytest.fixture()
def data(_data):
    return _data.copy()


@pytest.mark.parametrize('freq', [1, 7, 'M'])
def test_resample(data, freq):
    hour = 18
    r = iccas.resample(data, freq=1)
    expected_index = pd.date_range(
        data.index[0].replace(hour=hour, minute=0),
        data.index[-1].replace(hour=hour, minute=0),
        freq="D",
    )
    assert (r.index == expected_index).all(axis=None)
    assert not r.isna().any(axis=None)


@pytest.mark.parametrize(
    "x, expected",
    [
        ([1, 2, 3, 4, 5], [1, 2, 3, 4, 5]),
        ([1, 2, 5, 6, 4, 3], [1, 2, NA, NA, NA, 3]),
        ([7, 6, 1, 2, 4, 6, 5], [NA, NA, 1, 2, 4, NA, 5]),
        ([5, 3, 4, 4, 1], [NA, NA, NA, NA, 1]),
        ([1] * 5, [1] * 5),
    ],
)
def test_nullify_series_local_bumps(x, expected):
    actual = nullify_series_local_bumps(pd.Series(x)).fillna(NA).tolist()
    assert actual == expected



import datetime
from pathlib import Path
from typing import Optional, Tuple
from urllib.parse import urljoin

import pandas as pd
from iccas.caching import RemoteFolderProxy
from iccas.types import DateLike, PathType

BASE_URL = (
    "https://raw.githubusercontent.com/janLuke/iccas-dataset/master/data/"
)
AVAILABLE_FORMATS = {"csv"}
DEFAULT_CACHE_DIR = Path("~/.iccas").expanduser()
POPULATION_BY_AGE_PATH = "util/italian_population_by_age_2020.csv"
POPULATION_BY_AGE_GROUP_PATH = "util/italian_population_by_age_group_2020.csv"


def _check_file_format(fmt: str):
    if fmt not in AVAILABLE_FORMATS:
        raise ValueError(f"unavailable format: {fmt}")


def _date_as_str(date: DateLike) -> str:
    if isinstance(date, str):
        return date
    if isinstance(date, datetime.date):
        return date.strftime("%Y-%m-%d")
    raise TypeError


def _get_dataset_path(date: Optional[DateLike] = None, fmt: str = "csv") -> str:
    if date:
        date_str = _date_as_str(date)
        return "by-date/iccas_{}.{}".format(date_str, fmt)
    return "iccas_full.{}".format(fmt)


def get_url(date: Optional[DateLike] = None, fmt: str = "csv") -> str:
    """Returns the url of a dataset in a given format. If `date` is None,
    returns the URL of the full dataset."""
    _check_file_format(fmt)
    return urljoin(BASE_URL, _get_dataset_path(date=date, fmt=fmt))


def load(path: PathType) -> pd.DataFrame:
    dataset = pd.read_csv(path, index_col=("date", "age_group"), parse_dates=["date"])
    return dataset.unstack(level="age_group")


def load_single_date(path: PathType, keep_date: bool = False) -> pd.DataFrame:
    """
    Loads a dataset containing data for a single date.

    By default (`keep_date=False`), the `date` column is dropped and the
    datetime is stored in the `attrs` of the DataFrame.
    If instead `keep_date=True`, the returned dataset has a MultiIndex
    `(date, age_group)`.

    Args:
        path:
        keep_date: whether to drop the date column (containing a single datetime value)
    """
    index_col = ("date", "age_group") if keep_date else "age_group"
    dataset = pd.read_csv(path, index_col=index_col, parse_dates=["date"])
    if not keep_date:
        date = dataset.date.iloc[0]
        dataset = dataset.drop(columns="date")
        dataset.attrs["date"] = date
    return dataset


def get(cache_dir: PathType = DEFAULT_CACHE_DIR) -> pd.DataFrame:
    """
    Returns the latest version of the ICCAS dataset in a
    :class:`pandas.DataFrame` (as it's returned by :func:`load`).

    This function uses :meth:`RemoteFolderCache.get`, which caches.

    Raises:
        :exc:`request.exceptions.ConnectionError`: if the server is unreachable
        and no dataset is available in `cache_dir`
    """
    data_proxy = RemoteFolderProxy(folder_url=BASE_URL, local_path=cache_dir)
    path = data_proxy.get(_get_dataset_path())
    return load(path)


def get_by_date(
    date: DateLike,
    keep_date: bool = False,
    cache_dir: PathType = DEFAULT_CACHE_DIR,
) -> Tuple[pd.DataFrame, pd.Timestamp]:
    data_proxy = RemoteFolderProxy(folder_url=BASE_URL, local_path=cache_dir)
    path = data_proxy.get(_get_dataset_path(date=date))
    return load_single_date(path, keep_date=keep_date)


def get_population_by_age(cache_dir: PathType = DEFAULT_CACHE_DIR) -> pd.DataFrame:
    """
    Returns a DataFrame with "age" as index and two columns:
    "value" (absolute counts) and "percentage" (<=1.0)
    """
    data_proxy = RemoteFolderProxy(folder_url=BASE_URL, local_path=cache_dir)
    path = data_proxy.get(POPULATION_BY_AGE_PATH)
    pop = pd.read_csv(path, index_col=0)
    pop['percentage'] = pop.value / pop.value.sum()
    return pop


def get_population_by_age_group(
    cache_dir: PathType = DEFAULT_CACHE_DIR,
) -> pd.DataFrame:
    """
    Returns a DataFrame with "age_group" as index and two columns:
    "value" (absolute counts) and "percentage" (<=1.0)
    """
    data_proxy = RemoteFolderProxy(folder_url=BASE_URL, local_path=cache_dir)
    path = data_proxy.get(POPULATION_BY_AGE_GROUP_PATH)
    pop = pd.read_csv(path, index_col=0)
    pop['percentage'] = pop.value / pop.value.sum()
    return pop



from itertools import product
from typing import Dict, Iterable, List, Sequence, Set, Union

import pandas as pd

import iccas as ic
from iccas.charts.common import resample_if_needed
from iccas.types import PandasObj

PREFIX_MAP = {
    "T": "",
    "M": "male_",
    "F": "female_",
}
PREFIX_CODES = set(PREFIX_MAP)
FIELDS = ["cases", "cases_percentage", "deaths", "deaths_percentage", "fatality_rate"]
FIELD_SET = set(FIELDS)

AGE_GROUPS = (
    *(f"{start}-{start + 9}" for start in range(0, 90, 10)),
    ">=90",
    "unknown",
)


def product_join(*string_iterables, sep: str = "") -> Iterable[str]:
    return map(sep.join, product(*string_iterables))


def _check_column_name_part(values: Sequence[str], allowed_values: Set[str], kind: str):
    assert kind in {"prefixes", "fields"}
    if not values:
        raise ValueError(f'empty {kind} not allowed; use "*" to select all prefixes')
    if len(values) != len(set(values)):
        raise ValueError(f"invalid {kind}: duplicates not allowed")
    invalid_values = set(values) - allowed_values
    if invalid_values:
        raise ValueError(f"invalid {kind}: {invalid_values}")


def _check_prefix_codes(prefixes: Sequence[str]):
    _check_column_name_part(prefixes, PREFIX_CODES, "prefixes")


def _check_fields(fields: Sequence[str]):
    _check_column_name_part(fields, FIELD_SET, "fields")


def cols(prefixes: str, fields: Union[str, Sequence[str]] = "*") -> List[str]:
    """
    Generates a list of columns by combining prefixes with fields.

    Args:
        prefixes:
            string containing one or multiple of the following characters:
            - 'm' for males
            - 'f' for females
            - 't' for totals (no prefix)
            - '*' for all

        fields:
            values: 'cases', 'deaths', 'cases_percentage', 'deaths_percentage',
            'fatality_rate', '*'

    Returns:
        a list of string
    """
    field_list: Iterable[str]
    if isinstance(fields, str):
        if fields == "*":
            field_list = list(FIELDS)
        else:
            field_list = fields.split()
            _check_fields(field_list)
    else:
        field_list = fields
        _check_fields(field_list)

    prefix_list: Iterable[str]
    if prefixes == "*":
        prefix_list = PREFIX_MAP.values()
    else:
        prefixes = prefixes.upper()
        _check_prefix_codes(prefixes)
        prefix_list = [PREFIX_MAP[p] for p in prefixes]

    return list(product_join(prefix_list, field_list))


def only_counts(data: pd.DataFrame) -> pd.DataFrame:
    """Returns only cases and deaths columns (including sex-specific columns),
    dropping all other columns that are computable from these."""
    return data[
        [
            "cases",
            "female_cases",
            "male_cases",
            "deaths",
            "female_deaths",
            "male_deaths",
        ]
    ]


def only_cases(data: pd.DataFrame) -> pd.DataFrame:
    """ Returns only columns ['cases', 'female_cases', 'male_cases'] """
    return data[["cases", "female_cases", "male_cases"]]


def only_deaths(data: pd.DataFrame) -> pd.DataFrame:
    """ Returns only columns ['deaths', 'female_deaths', 'male_deaths'] """
    return data[["deaths", "female_deaths", "male_deaths"]]


def age_grouper(
    cuts: Union[int, Sequence[int]],
    fmt_last: str = ">={}",
) -> Dict[str, str]:
    if isinstance(cuts, int):
        if cuts % 10 != 0 or cuts < 0:
            raise ValueError()
        cuts = list(range(0, 100, cuts))
    else:
        for c in cuts:
            if c < 0 or c % 10 != 0:
                raise ValueError(f"at least one cut is not multiple of 10: {c}")
        if any(cuts[i - 1] >= cuts[i] for i in range(1, len(cuts))):
            raise ValueError(f"cuts  are not in increasing order: {cuts}")

    keys = AGE_GROUPS
    starts = [0, *cuts[:-1]]
    widths = [(end - start) // 10 for start, end in zip(starts, cuts)]
    grouper = {}

    k = 0
    for start, cut, width in zip(starts, cuts, widths):
        end = cut - 1
        value = f"{start}-{end}"
        for _ in range(width):
            grouper[keys[k]] = value
            k += 1
    last_value = fmt_last.format(cuts[-1])
    for _ in range(10 - len(grouper)):
        grouper[keys[k]] = last_value
        k += 1
    grouper["unknown"] = "unknown"
    return grouper


def aggregate_age_groups(
    counts: PandasObj,
    cuts: Union[int, Sequence[int]],
    fmt_last: str = ">={}",
) -> PandasObj:
    """
    Aggregates counts for different age groups summing them together.

    Args:
        counts:
            can be a Series with age groups as index or a DataFrame with
            age groups as columns, either in a simple Index or in
            a MultiIndex (no matter in what level)
        cuts:
            a single integer N means "cuts each N years";
            a sequence of integers determines the start ages of new age groups.
        fmt_last:
            format string for the last "unbounded" age group

    Returns:
        A Series/DataFrame with the same "structure" of the input but with
        aggregated age groups.
    """
    grouper = age_grouper(cuts=cuts, fmt_last=fmt_last)
    if isinstance(counts, pd.Series):
        return counts.groupby(grouper).sum()
    d = counts
    if "age_group" in counts.columns.names:
        d = counts.stack("age_group")
    d = d.reset_index("age_group")

    result = (
        d.assign(age_group=d.age_group.apply(grouper.__getitem__))
        .groupby(["date", "age_group"])
        .sum()
        .unstack("age_group")
    )

    if not isinstance(counts.columns, pd.MultiIndex):
        return result.droplevel(axis=1, level=0)
    return result


def get_unknown_sex_count(counts: pd.DataFrame, variable: str) -> pd.DataFrame:
    """ Returns cases/deaths of unknown sex for each age group """
    if variable not in {"cases", "deaths"}:
        raise ValueError("variable should be 'cases' or 'deaths'")
    total = counts[variable]
    sum_of_sexes = counts[f"male_{variable}"] + counts[f"female_{variable}"]
    return total - sum_of_sexes


def running_count(
    counts: PandasObj,
    window: int = 7,
    step: int = 1,
    **resample_kwargs
) -> PandasObj:
    """
    Given counts for cases and/or deaths, returns the number of new cases inside
    a temporal window of ``window`` days that moves forward by steps of ``step`` days.

    Args:
        counts:
        window:
        step:

    Returns:
    """
    if window % step != 0:
        raise ValueError("'window' must be a multiple of 'step'")
    r = ic.resample(counts, freq=step, **resample_kwargs)
    diff_periods = window // step
    out = r.diff(diff_periods).iloc[diff_periods:].round().astype(int)
    out.index = out.index.to_period(f'{window}D')
    return out


def running_average(
    counts: PandasObj,
    window: int = 7,
    step: int = 1,
    **resample_kwargs
) -> PandasObj:
    """
    Given counts for cases/deaths, returns the average daily number of
    new cases/deaths inside a temporal window of ``window``, moving the window
    ``step`` days a time.

    Args:
        counts:
        window:
        step:

    Returns:
    """
    return running_count(counts, window, step, **resample_kwargs) / window


def count_by_period(counts: PandasObj, freq: Union[str, int]) -> PandasObj:
    if isinstance(freq, int):
        freq = f'{freq}D'
    r = ic.resample(counts, freq=freq)
    r = r.diff().dropna()
    r.index = r.index.to_period(freq=freq).rename('period')
    return r


def average_by_period(counts: PandasObj, freq: Union[str, int]) -> PandasObj:
    c = count_by_period(counts, freq)
    lengths = (c.index.end_time - c.index.start_time + pd.Timedelta(nanoseconds=1)).days
    return c.div(lengths, axis=0)


def fatality_rate(counts, shift):
    """
    Computes the fatality rate as a ratio between the total number of deaths and
    the total number of cases ``shift`` days before.

    ``counts`` is resampled with interpolation if needed.
    """
    resampled = resample_if_needed(counts, freq='1D')
    shifted_cases = resampled.cases.shift(shift).iloc[shift:]
    deaths = resampled.deaths.iloc[shift:]
    cfr = (deaths / shifted_cases)
    if 'unknown' in cfr.columns:
        cfr = cfr.drop(columns='unknown')
    return cfr



__all__ = [
    "area",
    "bar",
    "area_chart",
    "double_area_chart_of_cumulative_counts",
    "double_area_chart_of_running_averages",
    "AgeDistributionBarChart",
    "age_dist_bar_chart",
    "average_by_period_bar_chart",
]
from . import area, bar
from .area import (
    area_chart,
    double_area_chart_of_cumulative_counts,
    double_area_chart_of_running_averages,
)
from .bar import (
    AgeDistributionBarChart,
    age_dist_bar_chart,
    average_by_period_bar_chart,
)



from typing import Optional

import matplotlib as mpl
import pandas as pd
from matplotlib import pyplot as plt

import iccas as ic
import iccas.i18n
from iccas.i18n import translated
from iccas.i18n.lib import Translation
from iccas.charts.common import (
    DEFAULT_CMAP,
    Period,
    check_age_group_size,
    check_variable,
    legend,
    resample_if_needed,
)

DOUBLE_CHART_FIGSIZE = (15, 15)


@translated()
def area_chart(
    counts_over_time: pd.DataFrame,
    *,
    strings: Translation,
    normalize: bool = False,
    ax: Optional[plt.Axes] = None,
    cmap=DEFAULT_CMAP,
    alpha: float = 0.8,
    **kwargs,
) -> plt.Axes:
    """
    A base for stacked area charts (tuned for iccas charts).

    Args:
        counts_over_time:
            a DataFrame with a DateTimeIndex and a column for each age group
            (e.g. ``ic.get().cases``)
        strings:
        normalize:
        ax:
        cmap:
        alpha:
        **kwargs:

    Returns:

    """
    data = counts_over_time
    if normalize:
        data = data.divide(data.sum(axis=1), axis=0)
    ax = data.plot.area(ax=ax, alpha=alpha, cmap=cmap, **kwargs)
    if normalize:
        ax.set_ylim([0, 1.0])
        ax.yaxis.set_major_formatter(mpl.ticker.PercentFormatter(xmax=1.0, decimals=0))
    else:
        ax.yaxis.set_major_formatter(mpl.ticker.StrMethodFormatter("{x:n}"))
    ax.set_xlabel("")
    ax.set_xlim(data.index[0], data.index[-1])
    legend(ax, title=strings["age"])
    return ax


@translated("""
    en:
        title.cases: Coronavirus cases in Italy since the beginning of the pandemic
        title.deaths: Deaths for coronavirus in Italy since the beginning of the pandemic
    it:
        title.cases: Casi di coronavirus dall'inizio della epidemia
        title.deaths: Decessi per coronavirus dall'inizio della pandemia
""")
def double_area_chart_of_cumulative_counts(
    data: pd.DataFrame,
    variable: str = "cases",
    *,
    age_group_size: int = 20,
    period: Optional[Period] = None,
    strings: Translation,
    **figure_args,
) -> plt.Figure:
    """
    Not a very interesting chart.

    Args:
        data:
            DataFrame having 'cases' and/or 'deaths' as first-level columns
            (e.g. ``ic.get()``)
        variable:
        age_group_size:
        period:
        strings:
        lang:

    Returns:

    """
    check_variable(variable)
    check_age_group_size(age_group_size)
    period_slice: slice = slice(*period) if period else slice(data.index[0], None)

    # resample for a smoother graph
    d = resample_if_needed(data[variable], "D", hour=18, method="pchip")
    d = ic.aggregate_age_groups(
        d[period_slice].drop(columns="unknown"),
        cuts=age_group_size,
        fmt_last="{}+",
    )

    figure_args.setdefault("figsize", DOUBLE_CHART_FIGSIZE)
    fig, ax = plt.subplots(2, 1, **figure_args)
    title = strings[f"title.{variable}"]
    ax[0].set_title(title)
    area_chart(d, ax=ax[0], lang=strings.lang)
    area_chart(d, normalize=True, ax=ax[1], lang=strings.lang)


@translated()
def double_area_chart_of_running_averages(
    data: pd.DataFrame,
    variable: str = "cases",
    *,
    strings: Translation,
    window: int = 14,
    age_group_size: int = 20,
    period: Optional[Period] = None,
    **figure_args,
):
    check_variable(variable)
    check_age_group_size(age_group_size)
    period_slice = slice(*period) if period else slice(data.index[0], None)

    r = resample_if_needed(data[variable].drop(columns="unknown"), "D")
    d = r[period_slice]
    d = d.diff(window).iloc[window:]
    d = ic.aggregate_age_groups(d, cuts=age_group_size, fmt_last="{}+")
    if window > 1:
        d = d / window

    figure_args.setdefault("figsize", DOUBLE_CHART_FIGSIZE)
    fig, ax = plt.subplots(2, 1, **figure_args)
    area_chart(d, ax=ax[0], lang=strings.lang)
    area_chart(d, ax=ax[1], normalize=True, lang=strings.lang)
    ax[0].set_title(strings.get(f"running_average_{variable}_title", count=window))
    return fig


if __name__ == "__main__":
    plt.style.use('seaborn')
    resampled = ic.resample(ic.get()[["cases"]])

    iccas.i18n.set_language('en')
    double_area_chart_of_running_averages(resampled, window=14)
    plt.show()

    iccas.i18n.set_language('it')
    double_area_chart_of_running_averages(resampled, window=14)
    plt.show()



import json
import os
from _warnings import warn
from pathlib import Path
from typing import Optional
from urllib.parse import unquote, urljoin

import requests

from iccas.types import PathType

METADATA_FNAME = ".meta.json"


class RemoteFolderProxy:
    def __init__(self, folder_url: str, local_path: PathType):
        """
        Mirrors a "remote folder" lazily downloading and caching its files.

        The caching mechanism uses only ETags (If-Modified-Since is not supported
        by raw.github).
        You should not modify files inside the cache folder; if you do, the file
        will be re-downloaded and overwritten at the next call of get().
        """
        self.folder_url = folder_url
        self.local_path = Path(local_path)
        self.local_path.mkdir(parents=True, exist_ok=True)
        self.metadata_path = self.local_path / METADATA_FNAME
        if not self.metadata_path.exists():
            _write_json(self.metadata_path, {})

    def get(self, relative_path, force_download: bool = False) -> Path:
        """
        Ensures the latest version of a remote file is available locally in the
        cache, downloading it only if needed.
        If no internet connection is available (or the server is unreachable),
        the file available in the cache is returned with a warning; if the file
        is not in the cache, a ``ConnectionError`` is raised.

        Args:
            relative_path:
            force_download:

        Returns:
            full local path of the file
        """
        path = self.get_path_of(relative_path)
        full_json = _read_json(self.metadata_path)
        entries = full_json[self.folder_url] if self.folder_url in full_json else {}
        if relative_path not in entries:
            entries[relative_path] = {}

        if (
            not force_download
            and path.exists()
            and entries[relative_path].get("creationTime") == os.path.getmtime(path)
        ):
            current_etag = entries[relative_path].get("ETag", "")
        else:
            current_etag = ""

        try:
            url = urljoin(self.folder_url, relative_path)
            resp = _download_if_modified(url, path, current_etag)
            if resp:
                entries[relative_path] = {
                    "ETag": resp.headers.get("ETag"),
                    "creationTime": os.path.getmtime(path),
                }
                full_json[self.folder_url] = entries
                _write_json(self.metadata_path, full_json)
        except requests.exceptions.ConnectionError as exc:
            if force_download:
                raise ConnectionError(
                    f"set force_download=False to use the local cached version"
                    f"of the file (if available).\nError details: {exc}"
                )
            if path.exists():
                warn(
                    f"Could not check if the remote file was modified: {exc}\n"
                    f"However, a file is available in the local cache."
                )
            else:
                raise ConnectionError(
                    "unable to download the dataset and no dataset was "
                    "previously downloaded.\nDetails: {}".format(exc)
                )
        return path

    def get_path_of(self, relative_url) -> Path:
        return Path(self.local_path, unquote(relative_url))


def _download_if_modified(
    url: str, path: PathType, etag: str
) -> Optional[requests.Response]:
    """ If the remote resource was modified, returns its new ETag, otherwise None. """
    path = Path(path)
    resp = requests.get(url, headers={"If-None-Match": etag})
    if resp.status_code == 304:
        return None
    else:
        resp.raise_for_status()
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "wb") as f:
            f.write(resp.content)
        return resp


def _read_json(path: PathType) -> dict:
    with open(path, encoding="utf-8") as f:
        return json.load(f)


def _write_json(path: PathType, data: dict):
    with open(path, "w", encoding="utf-8") as f:
        json.dump(data, f, indent=2)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "pandas",
    "requests",
    "numpy",
    "url-normalize",
    "pyyaml",
    "matplotlib",
    "seaborn",
]

extra_requirements = {
    'cli': ['cloup', 'click', 'tqdm', 'pathvalidate'],
}

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest>=3",
]

setup(
    author="Gianluca Gippetto",
    author_email="gianluca.gippetto@gmail.com",
    python_requires=">=3.7",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Helper library for the ICCAS dataset",
    install_requires=requirements,
    extras_require=extra_requirements,
    license="MIT license",
    long_description_content_type="text/x-rst",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="iccas",
    name="iccas",
    packages=find_packages(include=["iccas", "iccas.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/janLuke/iccas-python",
    version="0.1.2",
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'iccas=iccas.__main__:main',
        ],
    },
)
